from operator import mul
from functools import reduce
from fileformats.core import FileSet
from frametree.core.frameset.base import FrameSet


def test_column_api_access(dataset: FrameSet):

    bp = dataset.__annotations__["blueprint"]

    for fileset_bp in bp.entries:

        dataset.add_source(fileset_bp.path, fileset_bp.datatype)

        col = dataset[fileset_bp.path]

        # Check length of column
        assert len(col) == reduce(mul, bp.dim_lengths)

        # Access file-set via leaf IDs
        with dataset.tree:
            for id_ in col.ids:
                item = col[id_]
                assert isinstance(item, fileset_bp.datatype)
                if issubclass(fileset_bp.datatype, FileSet):
                    assert sorted(p.name for p in item.fspaths) == sorted(
                        fileset_bp.filenames
                    )



import operator as op
from itertools import chain
from functools import reduce, partial
import time
from multiprocessing import Pool, cpu_count
import pytest
from fileformats.generic import File
from fileformats.text import TextFile
from fileformats.field import Text as TextField
from frametree.core.frameset.base import FrameSet
from frametree.core.store import Store
from frametree.core.entry import DataEntry
from frametree.core.serialize import asdict
from frametree.common import FileSystem
from frametree.testing.blueprint import (
    TestDatasetBlueprint,
    FileSetEntryBlueprint as FileBP,
    FieldEntryBlueprint as FieldBP,
)
from frametree.testing import MockRemote


def test_populate_tree(dataset: FrameSet):
    blueprint = dataset.__annotations__["blueprint"]
    for freq in dataset.axes:
        # For all non-zero bases in the row_frequency, multiply the dim lengths
        # together to get the combined number of rows expected for that
        # row_frequency
        num_rows = reduce(
            op.mul, (ln for ln, b in zip(blueprint.dim_lengths, freq) if b), 1
        )
        assert (
            len(dataset.rows(freq)) == num_rows
        ), f"{freq} doesn't match {len(dataset.rows(freq))} vs {num_rows}"


def test_populate_row(dataset):
    blueprint = dataset.__annotations__["blueprint"]
    for row in dataset.rows("abcd"):
        if isinstance(dataset.store, FileSystem):
            expected_paths = sorted(
                chain(
                    (e.path for e in blueprint.entries if isinstance(e, FieldBP)),
                    *(e.filenames for e in blueprint.entries if isinstance(e, FileBP)),
                )
            )
        else:
            expected_paths = sorted(e.path for e in blueprint.entries)
        entry_paths = sorted(e.path for e in row.entries)
        assert entry_paths == expected_paths


def test_get(dataset: FrameSet):
    blueprint = dataset.__annotations__["blueprint"]
    for entry_bp in blueprint.entries:
        dataset.add_source(entry_bp.path, datatype=entry_bp.datatype)
    for row in dataset.rows(dataset.leaf_freq):
        for entry_bp in blueprint.entries:
            item = row[entry_bp.path]
            if item.is_fileset:
                item.trim_paths()
                assert sorted(p.name for p in item.fspaths) == sorted(
                    entry_bp.filenames
                )
            else:
                assert item.value == entry_bp.expected_value


def test_post(dataset: FrameSet):
    blueprint = dataset.__annotations__["blueprint"]

    def check_inserted():
        """Check that the inserted items are present in the dataset"""
        for deriv_bp in blueprint.derivatives:  # name, freq, datatype, _
            for row in dataset.rows(deriv_bp.row_frequency):
                cell = row.cell(deriv_bp.path, allow_empty=False)
                item = cell.item
                if item.is_fileset and isinstance(dataset.store, FileSystem):
                    assert item.fspath.relative_to(dataset.id)
                assert isinstance(item, deriv_bp.datatype)
                if deriv_bp.datatype.is_fileset:
                    assert item.hash_files() == all_checksums[deriv_bp.path]
                else:
                    assert item.primitive(item.value) == item.primitive(
                        deriv_bp.expected_value
                    )

    all_checksums = {}
    with dataset.tree:
        for deriv_bp in blueprint.derivatives:  # name, freq, datatype, files
            dataset.add_sink(
                name=deriv_bp.path,
                datatype=deriv_bp.datatype,
                row_frequency=deriv_bp.row_frequency,
            )
            test_file = deriv_bp.make_item()
            if deriv_bp.datatype.is_fileset:
                all_checksums[deriv_bp.path] = test_file.hash_files()
            # Test inserting the new item into the store
            with dataset.tree:
                for row in dataset.rows(deriv_bp.row_frequency):
                    row[deriv_bp.path] = test_file
        check_inserted()  # Check that cached objects have been updated
    check_inserted()  # Check that objects can be recreated from store


def test_grid_roundtrip(dataset: FrameSet):
    definition = asdict(dataset, omit=["store", "name"])
    definition["store-version"] = "1.0.0"

    data_store = dataset.store

    with data_store.connection:
        data_store.save_frameset_definition(
            dataset_id=dataset.id, definition=definition, name="test_dataset"
        )
        reloaded_definition = data_store.load_frameset_definition(
            dataset_id=dataset.id, name="test_dataset"
        )
    assert definition == reloaded_definition


# We use __file__ here as we just need any old file and can guarantee it exists
@pytest.mark.parametrize("datatype,value", [(File, __file__), (TextField, "value")])
def test_provenance_roundtrip(datatype: type, value: str, saved_dataset: FrameSet):
    provenance = {"a": 1, "b": [1, 2, 3], "c": {"x": True, "y": "foo", "z": "bar"}}
    data_store = saved_dataset.store

    with data_store.connection:
        entry = data_store.create_entry("provtest@", datatype, saved_dataset.root)
        data_store.put(datatype(value), entry)  # Create the entry first
        data_store.put_provenance(provenance, entry)  # Save the provenance
        reloaded_provenance = data_store.get_provenance(entry)  # reload the provenance
        assert provenance == reloaded_provenance


def test_singletons():
    standard = set(["file_system"])
    assert set(Store.singletons()) & standard == standard


@pytest.mark.skipif(
    condition=cpu_count() < 2, reason="Not enough cpus to run test with multiprocessing"
)
def test_delayed_download(
    delayed_mock_remote: MockRemote, simple_dataset_blueprint: TestDatasetBlueprint
):

    dataset_id = "delayed_download"
    dataset = simple_dataset_blueprint.make_dataset(delayed_mock_remote, dataset_id)
    entry = next(iter(dataset.rows())).entry("file1")

    delayed_mock_remote.clear_cache()

    worker = partial(
        delayed_download,
        entry,
    )
    with Pool(2) as p:
        try:
            no_offset, with_offset = p.map(worker, [0.0, 0.001])
        finally:
            p.close()  # Marks the pool as closed.
            p.join()  # Required to get the concurrency to show up in test coverage

    assert no_offset == "file1.txt"
    assert with_offset == "modified"


def delayed_download(entry: DataEntry, start_offset: float):
    # Set the downloads off at slightly different times
    time.sleep(start_offset)
    text_file = TextFile(entry.item)
    contents = text_file.contents
    if not start_offset:
        with open(text_file.fspath, "w") as f:
            f.write("modified")
    return contents



from operator import itemgetter
import click
from frametree.core.store import Store
from frametree.core.serialize import ClassResolver
from frametree.core.utils import get_home_dir
from frametree.core.exceptions import FrameTreeUsageError
from .base import cli


@cli.group()
def store():
    pass


@store.command(
    help="""Saves the details for a new data store in the configuration file ('~/.frametree/stores.yaml').

TYPE of the storage class, typically the name of the package the type is defined, e.g. 'xnat'
or 'bids'. More specific types can be given by using a colon to separate the package name and
type, e.g. frametree.xnat:XnatViaCS

NAME The name the store will be saved as in the configuration file. This is used to refer to the
store when using the frametree CLI.
"""
)
@click.argument("type")
@click.argument("name")
@click.option(
    "--server",
    "-s",
    default=None,
    help="The URI of the server to connect to (if applicable)",
)
@click.option(
    "--user", "-u", default=None, help="The username to use to connect to the store"
)
@click.option(
    "--password",
    "-p",
    prompt=True,
    hide_input=True,
    help="The password to use to connect to the store",
)
@click.option(
    "--cache",
    "-c",
    default=None,
    help="The location of a cache dir to download local copies of remote data",
)
@click.option(
    "--race-condition-delay",
    "-d",
    type=int,
    help=(
        "How long to wait for changes on a incomplete download before assuming it has "
        "been interrupted, clearing it and starting again"
    ),
)
@click.option(
    "--option",
    "-o",
    nargs=2,
    multiple=True,
    metavar="<name> <value>",
    default=None,
    help="Additional key-word arguments that are passed to the store class",
)
def add(name, type, option, cache, **kwargs):
    if option is not None:
        options = dict(option)
        conflicting = set(options) & set(kwargs)
        if conflicting:
            raise FrameTreeUsageError(
                f"Custom options {conflicting} conflict with in-built options, please "
                "use them instead"
            )
        kwargs.update(options)
    store_cls = ClassResolver(Store)(type)
    if hasattr(store_cls, "cache_dir"):
        if cache is None:
            cache = get_home_dir() / "cache" / name
        cache.mkdir(parents=True, exist_ok=True)
        kwargs["cache_dir"] = cache
    store = store_cls(name=name, **kwargs)
    store.save(name)


@store.command(
    help="""
Renames a store in the configuration file

OLD_KNAME The current name of the store.

NEW_NAME The new name for the store.
"""
)
@click.argument("old_name")
@click.argument("new_name")
def rename(old_name, new_name):
    Store.load(old_name).save(new_name)
    Store.remove(old_name)


@store.command(
    help="""Remove a saved data store from the config file

NAME The name the store was given when its details were saved
"""
)
@click.argument("name")
def remove(name):
    Store.remove(name)


@store.command(
    help="""Refreshes credentials saved for the given store (typically a token that expires)

NAME The name the store was given when its details were saved
"""
)
@click.argument("name")
@click.option(
    "--user", "-u", default=None, help="The username to use to connect to the store"
)
@click.option(
    "--password",
    "-p",
    prompt=True,
    hide_input=True,
    help="The password to use to connect to the store",
)
def refresh(name, user, password):
    store = Store.load(name)
    if user is not None:
        store.user = user
    store.password = password
    store.save()
    Store.remove(name)
    store.save(name)


@store.command(help="""List available stores that have been saved""")
def ls():
    click.echo("Default stores\n---------------")
    for name, store in sorted(Store.singletons().items(), key=itemgetter(0)):
        click.echo(f"{name} - {ClassResolver.tostr(store, strip_prefix=False)}")
    click.echo("\nSaved stores\n-------------")
    for name, entry in Store.load_saved_configs().items():
        store_class = entry.pop("class")
        click.echo(f"{name} - {store_class[1:-1]}")
        for key, val in sorted(entry.items(), key=itemgetter(0)):
            click.echo(f"    {key}: {val}")



from operator import itemgetter
import pytest
from frametree.core.store import Store
from frametree.core.exceptions import FrameTreeConstructionError


ID_INFERENCE_TESTS = {
    "basic": (
        {"abcd": "a1b2c3d4"},
        {
            "a": r"abcd::a(\d+)b\d+c\d+d\d+",
            "b": r"abcd::a\d+b(\d+)c\d+d\d+",
            "c": r"abcd::a\d+b\d+c(\d+)d\d+",
            "d": r"abcd::a\d+b\d+c\d+d(\d+)",
        },
        {},
        {
            "a": "1",
            "b": "2",
            "c": "3",
            "d": "4",
        },
    ),
    "realistic": (
        {"subject": "CONTROL99"},
        {
            "group": r"subject:id:([A-Z]+).*",
            "member": r"subject:ID:[A-Z]+(\d+)",
        },
        {},
        {
            "group": "CONTROL",
            "member": "99",
        },
    ),
    "metadata": (
        {"subject": "FTD01111", "session": "FTDS9999"},
        {
            "visit": r"session:order",
        },
        {"session": {"order": 2}},
        {
            "visit": "2",
        },
    ),
    "templating": (
        {"subject": "FTD01111", "session": "MR999"},
        {
            "visit": r"#session::([A-Z]+).*##session:visit_id#",
        },
        {"session": {"visit_id": 3}},
        {
            "visit": "MR3",
        },
    ),
}


@pytest.mark.parametrize(
    "fixture",
    ID_INFERENCE_TESTS.items(),
    ids=itemgetter(0),
)
def test_id_inference(fixture):
    test_name, (explicit_ids, id_patterns, metadata, expected_ids) = fixture
    assert expected_ids == Store.infer_ids(
        ids=explicit_ids, id_patterns=id_patterns, metadata=metadata
    )


def test_id_inference_fail1():
    with pytest.raises(
        FrameTreeConstructionError, match="Provided ID-pattern component"
    ):
        Store.infer_ids(ids={"ab": "a0b0"}, id_patterns={"a": r"ab::xxx"})


def test_id_inference_fail2():
    with pytest.raises(
        FrameTreeConstructionError, match="Provided ID-pattern component"
    ):
        Store.infer_ids(ids={"ab": "a0b0"}, id_patterns={"a": r"ab::a\d+b\d+"})


def test_id_inference_fail3():
    with pytest.raises(
        FrameTreeConstructionError, match="Provided ID-pattern component"
    ):
        Store.infer_ids(ids={"ab": "a0b0"}, id_patterns={"a": r"ab::a(\d+)b(\d+)"})


def test_id_inference_fail4():
    with pytest.raises(
        FrameTreeConstructionError, match="row doesn't have the metadata field"
    ):
        Store.infer_ids(ids={"ab": "a0b0"}, id_patterns={"a": r"ab:metadata_field"})


def test_id_inference_fail5():
    with pytest.raises(
        FrameTreeConstructionError,
        match="Inferred IDs from decomposition conflict",
    ):
        Store.infer_ids(ids={"ab": "a0b0"}, id_patterns={"ab": r"ab::a(\d+)b\d+"})



from __future__ import annotations
import logging
import typing as ty
import re
from collections import defaultdict
import attrs
import attrs.filters
from frametree.core.utils import NestedContext
from frametree.core.axes import Axes
from frametree.core.exceptions import (
    FrameTreeNameError,
    FrameTreeConstructionError,
)
from .row import DataRow

if ty.TYPE_CHECKING:  # pragma: no cover
    from .frameset.base import FrameSet


logger = logging.getLogger("frametree")


def auto_ids_default():
    return defaultdict(dict)


@attrs.define
class DataTree(NestedContext):

    frameset: ty.Optional[FrameSet] = None
    root: ty.Optional[DataRow] = None
    _auto_ids: ty.Dict[ty.Tuple[str, ...], ty.Dict[str, int]] = attrs.field(
        factory=auto_ids_default
    )

    def enter(self):
        assert self.root is None
        self._set_root()
        self.frameset.store.populate_tree(self)

    def exit(self):
        self.root = None

    @property
    def dataset_id(self):
        return self.frameset.id

    @property
    def hierarchy(self):
        return self.frameset.hierarchy

    def add_leaf(
        self, tree_path, metadata: ty.Dict[str, ty.Dict[str, str]] = None
    ) -> ty.Tuple[DataRow, ty.List[str]]:
        """Creates a new row at a the path down the tree of the dataset as
        well as all "parent" rows upstream in the data tree

        Parameters
        ----------
        tree_path : list[str]
            The sequence of labels for each layer in the hierarchy of the
            dataset leading to the current row.
        metadata : dict[str, ty.Dict[str, str]]
            metadata passed to ``Store.infer_ids()`` used to infer IDs not directly
            represented in the hierarchy of the data tree.

        Returns
        -------
        row : DataRow or None
            the added row if it is not excluded, None if it was excluded
        exclusions : list[str]
            the list of frequencies that caused the leaf to be excluded (empty if it
            was added) according to the the exclusion criteria provided to the dataset
            (see ``FrameSet.include`` and ``FrameSet.exclude``)

        Raises
        ------
        FrameTreeBadlyFormattedIDError
            raised if one of the IDs doesn't match the pattern in the
            `id_patterns`
        FrameTreeConstructionError
            raised if one of the groups specified in the ID inference reg-ex
            doesn't match a valid row_frequency in the data dimensions
        """

        def matches_criteria(
            label: str, freq_str: str, criteria: ty.Dict[str, ty.Union[list, str]]
        ):
            try:
                freq_criteria = criteria[freq_str]
            except KeyError:
                return None
            if isinstance(freq_criteria, list):
                return label in freq_criteria
            else:
                return bool(re.match(freq_criteria, label))

        if self.root is None:
            self._set_root()
        if metadata is None:
            metadata = {}
        # Get basis frequencies covered at the given depth of the
        if len(tree_path) != len(self.frameset.hierarchy):
            raise FrameTreeConstructionError(
                f"Tree path ({tree_path}) should have the same length as "
                f"the hierarchy ({self.frameset.hierarchy}) of {self}"
            )
        if self.frameset.exclude:
            for freq_str, label in zip(self.frameset.hierarchy, tree_path):
                if matches_criteria(label, freq_str, self.frameset.exclude):
                    return None  # Don't add leaf
        # Set a default ID of None for all parent frequencies that could be
        # inferred from a row at this depth
        # ids = {f: None for f in self.frameset.axes}
        ids = dict(zip(self.frameset.hierarchy, tree_path))
        # Infer IDs and add them to those explicitly in the hierarchy
        inferred_ids = self.frameset.infer_ids(ids, metadata=metadata)
        ids.update(inferred_ids)
        # Calculate the combined freqs after each layer is added
        cummulative_freq = self.frameset.axes(0)
        for i, layer_str in enumerate(self.frameset.hierarchy):
            layer_freq = self.frameset.axes[layer_str]
            # If all the axes introduced by the layer not present in parent layers
            # and none of the IDs of these axes have been inferred from other IDs,
            # then the ID of the axis out of the layer's axes with the least-
            # significant bit can be considered to be equivalent to the
            # ID of the layer and the IDs of the other axes of the layer set to None
            # (the order of # the bits in the Axes class should be arranged to
            # account for this default behaviour).
            #
            # For example, given a hierarchy of ['subject', 'session'] in the `Clinical`
            # data space, no groups are assumed to be present by default (i.e. if not
            # specified by the `id_patterns` attr of the dataset), and the `member`
            # ID is assumed to be equivalent to the `subject` ID, since `member`
            # correspdonds to the least significant bit in the value of the subject in
            # the `Clinical` data space enum.
            #
            # Conversely, the visit can't be assumed to be equal to the `session`
            # ID, since the session ID could be expected to also contain both the `member` and
            # `group` ID in it, and should be explicitly extracted by via `id_patterns`
            #
            #       session ID: MRH010_CONTROL03_MR02
            #
            # with the '02' part representing as the visit can be extracted with the
            #
            #       id_inference = {
            #           'visit': r'session:id:.*MR(0-9+)$'
            #       }
            # Axes already added by predecessor layers
            prev_accounted_for = layer_freq & cummulative_freq
            # Axes added by this layer
            new = prev_accounted_for ^ layer_freq
            assert new, f"{layer_str} doesn't add any new axes on predecessor layers"
            layer_span = [str(f) for f in layer_freq.span()]
            # Axes that have an ID already
            unresolved_axes = [f for f in layer_span if f not in ids]
            if unresolved_axes:
                for axis in unresolved_axes[:-1]:
                    ids[axis] = None
                # If all axes added by the layer are new and none are resolved to IDs
                # we can just use the ID for the layer to be equivalent to the last axis
                if not prev_accounted_for and unresolved_axes == layer_span:
                    assumed_id = ids[layer_str]
                else:
                    node_path = tuple(tree_path[:i]) + tuple(
                        ids[str(f)] for f in new.span() if str(f) in ids
                    )
                    layer_label = tree_path[i]
                    try:
                        assumed_id = self._auto_ids[node_path][layer_label]
                    except KeyError:
                        assumed_id = str(len(self._auto_ids[node_path]) + 1)
                        self._auto_ids[node_path][layer_label] = assumed_id
                ids[unresolved_axes[-1]] = assumed_id
            cummulative_freq |= layer_freq
        assert cummulative_freq == self.frameset.axes.leaf()
        assert set(ids).issuperset(str(f) for f in self.frameset.axes.bases())
        # # Set or override any inferred IDs within the ones that have been
        # # explicitly provided
        # clashing_ids = set(ids) & set(additional_ids)
        # if clashing_ids:
        #     raise FrameTreeUsageError(
        #         f"Additional IDs clash with those inferred: {clashing_ids}"
        #     )
        # ids.update(additional_ids)
        # Create composite IDs for non-basis frequencies if they are not
        # explicitly in the layer dimensions
        for freq in set(self.frameset.axes) - set(self.frameset.axes.bases()):
            freq_str = str(freq)
            if freq_str not in ids:
                id_ = tuple(ids[str(b)] for b in freq.span() if ids[str(b)] is not None)
                if id_:
                    if len(id_) == 1:
                        id_ = id_[0]
                else:
                    id_ = None
                ids[freq_str] = id_
        # Determine whether leaf node is included in the dataset definition according
        # to the include and exclude criteria
        if self.frameset.include:
            for freq in self.frameset.axes:
                freq_str = str(freq)
                if (
                    matches_criteria(ids[freq_str], freq_str, self.frameset.include)
                    is False
                ):
                    return None
        return self._add_row(
            ids={f: ids.get(str(f)) for f in self.frameset.axes},
            row_frequency=self.frameset.axes.leaf(),
        )

    def _add_row(self, ids: ty.Dict[Axes, str], row_frequency):
        """Adds a row to the dataset, creating all parent "aggregate" rows
        (e.g. for each subject, group or visit) where required

        Parameters
        ----------
        ids : dict[Axes, str]
            ids of the row in all frequencies that it intersects
        row: DataRow
            The row to add into the data tree

        Raises
        ------
        FrameTreeConstructionError
            If inserting a multiple IDs of the same class within the tree if
            one of their ids is None
        """
        # logger.debug(
        #     "Found %s row in %s dataset: %s", row_frequency, self.dataset_id, ids
        # )
        row_frequency = self.frameset.parse_frequency(row_frequency)
        row = DataRow(ids=ids, frequency=row_frequency, frameset=self.frameset)
        # Create new data row
        try:
            row_dict = self.root.children[row.frequency]
        except KeyError:
            row_dict = self.root.children[row.frequency] = {}
        if row.id in row_dict:
            raise FrameTreeConstructionError(
                f"ID clash ({row.id}) between rows inserted into the data tree of "
                f"{self.frameset.id} in {self.frameset.store.name} store:\n"
                "  exist: "
                + ", ".join(f"{f}={i}" for f, i in sorted(row_dict[row.id].ids.items()))
                + "\n  added: "
                + ", ".join(f"{f}={i}" for f, i in sorted(row.ids.items()))
            )
        row_dict[row.id] = row
        # Insert root row
        # Insert parent rows if not already present and link them with
        # inserted row
        for parent_freq, parent_id in row.ids.items():
            if not parent_freq:
                continue  # Don't need to insert root row again
            diff_freq = (row.frequency ^ parent_freq) & row.frequency
            if diff_freq:
                try:
                    parent_row = self.frameset.row(frequency=parent_freq, id=parent_id)
                except FrameTreeNameError:
                    parent_ids = {
                        f: i
                        for f, i in row.ids.items()
                        if f.is_parent(parent_freq, if_match=True)
                    }
                    parent_row = self._add_row(parent_ids, parent_freq)
                # Set reference to level row in new row
                diff_id = row.frequency_id(diff_freq)
                try:
                    children_dict = parent_row.children[row_frequency]
                except KeyError:
                    children_dict = parent_row.children[row_frequency] = {}
                if diff_id in children_dict:
                    raise FrameTreeConstructionError(
                        f"ID clash between rows inserted into data tree, {diff_id}, "
                        f"in {diff_freq} children of {parent_row} "
                        f"({children_dict[diff_id]} and {row}). You may "
                        f"need to set the `id_patterns` attr of the dataset "
                        "to disambiguate ID components (e.g. how to extract "
                        "the visit ID from a session label)"
                    )
                children_dict[diff_id] = row
        return row

    def _set_root(self):
        self.root = DataRow(
            ids={self.frameset.root_freq: None},
            frequency=self.frameset.root_freq,
            frameset=self.frameset,
        )
        self._auto_ids = auto_ids_default()



from ._version import __version__  # noqa: F401

PACKAGE_NAME = "frametree"
CODE_URL = f"https://github.com/ArcanaFramework/{PACKAGE_NAME}"

__authors__ = [("Thomas G. Close", "tom.g.close@gmail.com")]

from .frameset import FrameSet  # noqa: F401



from __future__ import annotations
import typing as ty
import attrs
from frametree.core.exceptions import (
    FrameTreeNameError,
    FrameTreeWrongFrequencyError,
)
from fileformats.core import DataType
from .quality import DataQuality
from .axes import Axes
from .cell import DataCell
from .entry import DataEntry


if ty.TYPE_CHECKING:  # pragma: no cover
    from .frameset.base import FrameSet


@attrs.define(kw_only=True)
class DataRow:
    """A "row" in a dataset "frame" where file-sets and fields can be placed, e.g.
    a session or subject.

    Parameters
    ----------
    ids : Dict[Axes, str]
        The ids for the frequency of the row and all "parent" frequencies
        within the tree
    dataset : FrameSet
        A reference to the root of the data tree
    frequency : str
        The frequency of the row
    tree_path : list[str], optional
        the path to the row within the data tree. None if the row doesn't sit within
        the original tree (e.g. visits within a subject>session hierarchy)
    uri : str, optional
        a URI for the row, can be set and used by the data store implementation if
        appropriate, by default None
    """

    ids: ty.Dict[Axes, str] = attrs.field()
    frameset: FrameSet = attrs.field(repr=False)
    frequency: str = attrs.field()
    tree_path: ty.List[str] = None
    uri: ty.Optional[str] = None
    metadata: ty.Optional[dict] = None

    # Automatically populated fields
    children: ty.Dict[Axes, ty.Dict[ty.Union[str, ty.Tuple[str]], str]] = attrs.field(
        factory=dict, repr=False, init=False
    )
    _entries_dict: ty.Dict[str, DataEntry] = attrs.field(
        default=None, init=False, repr=False
    )
    _cells: ty.Dict[str, DataCell] = attrs.field(factory=dict, init=False, repr=False)

    @frameset.validator
    def dataset_validator(self, _, dataset):
        from .frameset import FrameSet

        if not isinstance(dataset, FrameSet):
            raise ValueError(f"provided dataset {dataset} is not of type {FrameSet}")

    @frequency.validator
    def frequency_validator(self, _, frequency):
        if frequency not in self.frameset.axes:
            raise ValueError(
                f"'{frequency}' frequency is not in the data space of the dataset, "
                f"{self.frameset.axes}"
            )

    def __attrs_post_init__(self):
        if isinstance(self.frequency, str):
            self.frequency = self.frameset.axes[self.frequency]

    def __getitem__(self, column_name: str) -> DataType:
        """Gets the item for the current row

        Parameters
        ----------
        column_name : str
            Name of a selected column in the dataset

        Returns
        -------
        DataType
            The item matching the provided name specified by the column name
        """
        return self.cell(column_name).item

    def __setitem__(self, column_name: str, value: DataType):
        self.cell(column_name).item = value
        return self

    def cell(self, column_name: str, allow_empty: ty.Optional[bool] = None) -> DataCell:
        try:
            cell = self._cells[column_name]
        except KeyError:
            pass
        else:
            if not cell.is_empty:
                return cell
        try:
            column = self.frameset[column_name]
        except KeyError as e:
            raise FrameTreeNameError(
                column_name,
                f"{column_name} is not the name of a column in "
                f"{self.frameset.id} dataset ('"
                + "', '".join(self.frameset.columns)
                + "')",
            ) from e
        if column.row_frequency != self.frequency:
            return FrameTreeWrongFrequencyError(
                column_name,
                f"'column_name' ({column_name}) is of {column.row_frequency} "
                f"frequency and therefore not in rows of {self.frequency}"
                " frequency",
            )
        cell = DataCell.intersection(column=column, row=self, allow_empty=allow_empty)
        self._cells[column_name] = cell
        return cell

    def cells(self, allow_empty: ty.Optional[bool] = None) -> ty.Iterable[DataCell]:
        for column_name in self.frameset.columns:
            yield self.cell(column_name, allow_empty=allow_empty)

    @property
    def entries(self) -> ty.Iterable[DataEntry]:
        return self.entries_dict.values()

    def entry(self, id: str) -> DataEntry:
        return self.entries_dict[id]

    @property
    def entries_dict(self):
        if self._entries_dict is None:
            self._entries_dict = {}
            self.frameset.store.populate_row(self)
        return self._entries_dict

    def __repr__(self):
        return f"{type(self).__name__}(id={self.id}, frequency={self.frequency})"

    @property
    def id(self):
        return self.ids[self.frequency]

    @property
    def ids_tuple(self):
        return tuple(self.ids[a] for a in self.frameset.axes.axes())

    @property
    def label(self):
        return self.tree_path[-1]

    def frequency_id(self, frequency: ty.Union[str, Axes]):
        return self.ids[self.frameset.axes[str(frequency)]]

    def __iter__(self):
        return iter(self.keys())

    def keys(self):
        return (n for n, _ in self.items())

    def values(self):
        return (i for _, i in self.items())

    def items(self):
        return (
            (c.name, self[c.name])
            for c in self.frameset.columns.values()
            if c.row_frequency == self.frequency
        )

    def column_items(self, column_name):
        """Gets the item for the current row if item's frequency matches
        otherwise gets all the items that are related to the current row (
        i.e. are in child rows)

        Parameters
        ----------
        column_name : str
            Name of a selected column in the dataset

        Returns
        -------
        Sequence[DataType]
            The item matching the provided name specified by the column name
            if the column is of matching or ancestor frequency, or list of
            items if a descendent or unrelated frequency.
        """
        try:
            return [self[column_name]]
        except FrameTreeWrongFrequencyError:
            # If frequency is not a ancestor row then return the
            # items in the children of the row (if they are child
            # rows) or the whole dataset
            spec = self.frameset.columns[column_name]
            try:
                return self.children[spec.row_frequency].values()
            except KeyError:
                return self.frameset.column(spec.row_frequency)

    def add_entry(
        self,
        path: str,
        datatype: type,
        uri: str,
        item_metadata: ty.Optional[dict] = None,
        order: ty.Optional[int] = None,
        quality: DataQuality = DataQuality.usable,
        checksums: ty.Dict[str, str] = None,
    ):
        """Adds an data entry to a row that has been found while scanning the row in the
        repository.

        Parameters
        ----------
        id : str
            the ID of the entry within the node
        datatype : type (subclass of DataType)
            the type of the data entry
        uri : str
            a URI uniquely identifying the data entry, which can be used by the store
            for convenient/efficient access to the entry. Note that newly added entries
            (i.e. created by data-sink columns) will not have their URI set, so data
            store logic should fallback to using the row+id to identify the entry in
            the dataset.
        item_metadata : dict[str, Any]
            metadata associated with the data item itself (e.g. pulled from a file header).
            Can be supplied either when the entry is initialised (i.e. from previously
            extracted fields stored within the data store), or read from the item itself.
        order : int, optional
            the order in which the entry appears in the node (where applicable)
        provenance : dict, optional
            the provenance associated with the derivation of the entry by FrameTree
            (only applicable to derivatives not source data)
        checksums : dict[str, str], optional
            checksums for all of the files in the data entry
        """
        if self._entries_dict is None:
            self._entries_dict = {}
        entry = DataEntry(
            path=path,
            datatype=datatype,
            row=self,
            uri=uri,
            item_metadata=item_metadata,
            order=order,
            quality=quality,
            checksums=checksums,
        )
        if path in self._entries_dict:
            raise KeyError(
                f"Attempting to add multiple entries with the same path, '{path}', to "
                f"{self}, {self._entries_dict[path]} and {entry}"
            )
        self._entries_dict[path] = entry
        return entry



from __future__ import annotations
from operator import itemgetter
import pytest
import typing as ty
from fileformats.text import TextFile
from frametree.core.exceptions import FrameTreeUsageError
from frametree.common import FileSystem, Clinical
from frametree.testing.blueprint import TestDatasetBlueprint, FileSetEntryBlueprint
from frametree.testing.axes import TestAxes


TEST_INCLUSIONS = {
    "all": (
        TestDatasetBlueprint(  # dataset name
            axes=TestAxes,
            hierarchy=["a", "b", "c", "abcd"],
            dim_lengths=[1, 2, 3, 4],
            entries=[
                FileSetEntryBlueprint(
                    path="file1", datatype=TextFile, filenames=["file1.txt"]
                ),
            ],
            id_patterns={"d": r"abcd::.*(d\d+)"},
        ),
        [
            "a0b0c0d0",
            "a0b0c0d1",
            "a0b0c0d2",
            "a0b0c0d3",
            "a0b0c1d0",
            "a0b0c1d1",
            "a0b0c1d2",
            "a0b0c1d3",
            "a0b0c2d0",
            "a0b0c2d1",
            "a0b0c2d2",
            "a0b0c2d3",
            "a0b1c0d0",
            "a0b1c0d1",
            "a0b1c0d2",
            "a0b1c0d3",
            "a0b1c1d0",
            "a0b1c1d1",
            "a0b1c1d2",
            "a0b1c1d3",
            "a0b1c2d0",
            "a0b1c2d1",
            "a0b1c2d2",
            "a0b1c2d3",
        ],
    ),
    "include": (
        TestDatasetBlueprint(  # dataset name
            axes=TestAxes,
            hierarchy=["a", "b", "c", "abcd"],
            dim_lengths=[1, 2, 3, 4],
            entries=[
                FileSetEntryBlueprint(
                    path="file1", datatype=TextFile, filenames=["file1.txt"]
                ),
            ],
            include={
                "c": ["c1", "c2"],
                "a": "a0",
                "d": ["d1", "d2", "d3"],
            },
            id_patterns={"d": r"abcd::.*(d\d+)"},
        ),
        [
            "a0b0c1d1",
            "a0b0c1d2",
            "a0b0c1d3",
            "a0b0c2d1",
            "a0b0c2d2",
            "a0b0c2d3",
            "a0b1c1d1",
            "a0b1c1d2",
            "a0b1c1d3",
            "a0b1c2d1",
            "a0b1c2d2",
            "a0b1c2d3",
        ],
    ),
    "exclude": (
        TestDatasetBlueprint(  # dataset name
            axes=TestAxes,
            hierarchy=["a", "b", "c", "abcd"],
            dim_lengths=[1, 2, 3, 4],
            entries=[
                FileSetEntryBlueprint(
                    path="file1", datatype=TextFile, filenames=["file1.txt"]
                ),
            ],
            exclude={
                "c": ["c2"],
                "abcd": ["a0b0c1d1", "a0b1c1d2"],
            },
            id_patterns={"d": r"abcd::.*(d\d+)"},
        ),
        [
            "a0b0c0d0",
            "a0b0c0d1",
            "a0b0c0d2",
            "a0b0c0d3",
            "a0b0c1d0",
            "a0b0c1d2",
            "a0b0c1d3",
            "a0b1c0d0",
            "a0b1c0d1",
            "a0b1c0d2",
            "a0b1c0d3",
            "a0b1c1d0",
            "a0b1c1d1",
            "a0b1c1d3",
        ],
    ),
    "regex": (
        TestDatasetBlueprint(  # dataset name
            axes=TestAxes,
            hierarchy=["a", "b", "c", "abcd"],
            dim_lengths=[1, 2, 3, 4],
            entries=[
                FileSetEntryBlueprint(
                    path="file1", datatype=TextFile, filenames=["file1.txt"]
                ),
            ],
            id_patterns={"d": r"abcd::.*(d\d+)"},
            include={
                "abcd": r"a\d+b\dc(\d+)d\1",
            },
        ),
        [
            "a0b0c0d0",
            "a0b0c1d1",
            "a0b0c2d2",
            "a0b1c0d0",
            "a0b1c1d1",
            "a0b1c2d2",
        ],
    ),
}


@pytest.mark.parametrize("fixture", TEST_INCLUSIONS.items(), ids=itemgetter(0))
def test_dataset_inclusion(
    fixture: ty.Tuple[str, ty.Tuple[TestDatasetBlueprint, ty.List[str]]], work_dir
):
    test_name, (blueprint, expected) = fixture
    dataset_path = work_dir / test_name
    dataset = blueprint.make_dataset(store=FileSystem(), dataset_id=dataset_path)
    assert sorted(dataset.row_ids()) == expected


def test_include_exclude_fail1(work_dir):

    blueprint = TestDatasetBlueprint(  # dataset name
        axes=TestAxes,
        hierarchy=["a", "b", "c", "abcd"],
        dim_lengths=[1, 2, 3, 4],
        entries=[
            FileSetEntryBlueprint(
                path="file1", datatype=TextFile, filenames=["file1.txt"]
            ),
        ],
        include={"session": r"a0.*d1.*"},
    )
    with pytest.raises(
        FrameTreeUsageError, match="Unrecognised frequencies in 'include'"
    ):
        blueprint.make_dataset(
            store=FileSystem(), dataset_id=work_dir / "include-exclude-fail1"
        )


def test_include_exclude_fail2(work_dir):

    blueprint = TestDatasetBlueprint(  # dataset name
        axes=TestAxes,
        hierarchy=["a", "b", "c", "abcd"],
        dim_lengths=[1, 2, 3, 4],
        entries=[
            FileSetEntryBlueprint(
                path="file1", datatype=TextFile, filenames=["file1.txt"]
            ),
        ],
        exclude={"bc": r"a0.*d1.*"},
    )
    with pytest.raises(
        FrameTreeUsageError,
        match="only frequencies present in the dataset hierarchy are allowed",
    ):
        blueprint.make_dataset(
            store=FileSystem(), dataset_id=work_dir / "include-exclude-fail2"
        )


def test_include_exclude_fail3(work_dir):

    blueprint = TestDatasetBlueprint(  # dataset name
        axes=TestAxes,
        hierarchy=["a", "b", "c", "abcd"],
        dim_lengths=[1, 2, 3, 4],
        entries=[
            FileSetEntryBlueprint(
                path="file1", datatype=TextFile, filenames=["file1.txt"]
            ),
        ],
        include={"abcd": r"a**"},
    )
    with pytest.raises(FrameTreeUsageError, match="valid regular expression"):
        blueprint.make_dataset(
            store=FileSystem(), dataset_id=work_dir / "include-exclude-fail3"
        )


TEST_AUTO_IDS = {
    "d_dim": (
        TestDatasetBlueprint(  # dataset name
            axes=TestAxes,
            hierarchy=["a", "b", "c", "abcd"],
            dim_lengths=[1, 2, 3, 4],
            entries=[
                FileSetEntryBlueprint(
                    path="file1", datatype=TextFile, filenames=["file1.txt"]
                ),
            ],
        ),
        {
            "d": ["1", "2", "3", "4"],
        },
    ),
    "member": (
        TestDatasetBlueprint(  # dataset name
            axes=Clinical,
            hierarchy=["group", "subject", "visit"],
            dim_lengths=[2, 2, 2],
            entries=[
                FileSetEntryBlueprint(
                    path="file1", datatype=TextFile, filenames=["file1.txt"]
                ),
            ],
        ),
        {
            "member": ["1", "2"],
        },
    ),
    "double_increment": (
        TestDatasetBlueprint(  # dataset name
            axes=TestAxes,
            hierarchy=["ab", "cd"],
            dim_lengths=[2, 2, 2, 2],
            entries=[
                FileSetEntryBlueprint(
                    path="file1", datatype=TextFile, filenames=["file1.txt"]
                ),
            ],
            id_patterns={
                "b": r"ab::a\d+(b\d+)",
                "c": r"cd::(c\d+)d\d+",
            },
        ),
        {
            "a": ["1", "2"],
            "b": ["b0", "b1"],
            "c": ["c0", "c1"],
            "d": ["1", "2"],
        },
    ),
}


@pytest.mark.parametrize("fixture", TEST_AUTO_IDS.items(), ids=itemgetter(0))
def test_auto_ids(work_dir, fixture):

    _, (blueprint, expected) = fixture

    dataset = blueprint.make_dataset(
        store=FileSystem(), dataset_id=work_dir / "incrementing-ids"
    )

    for key, ids in expected.items():
        assert sorted(dataset.row_ids(key)) == ids



from fileformats.generic import Directory
from frametree.common.tasks import shell


def test_shell(work_dir):

    cp = shell(
        name="copy",
        executable="cp",
        inputs=[
            {
                "name": "in_dir",
                "datatype": "generic/directory",
            }
        ],
        outputs=[
            {
                "name": "out_dir",
                "datatype": "generic/directory",
                "position": -1,
            }
        ],
        parameters=[
            {
                "name": "recursive",
                "datatype": "field/boolean",
                "argstr": "-R",
                "position": 0,
            }
        ],
    )

    in_dir = work_dir / "source-dir"
    in_dir.mkdir()
    with open(in_dir / "a-file.txt", "w") as f:
        f.write("abcdefg")

    out_dir = work_dir / "dest-dir"

    result = cp(
        in_dir=str(in_dir),
        out_dir=str(out_dir),
        recursive=True,
    )

    assert result.output.out_dir == Directory(out_dir)
    assert list(p.name for p in out_dir.iterdir()) == ["a-file.txt"]



from functools import reduce
from operator import mul
from frametree.core.utils import show_cli_trace
from fileformats.text import TextFile
from pathlib import Path
import tempfile
import pytest
from frametree.core.frameset import FrameSet
from frametree.core.cli.processing import install_license
from frametree.core.cli.processing import derive, apply
from frametree.testing import MockRemote


def test_apply_cli(saved_dataset: FrameSet, concatenate_task, cli_runner):
    # Get CLI name for dataset (i.e. file system path prepended by 'file_system//')
    # Start generating the arguments for the CLI
    # Add source to loaded dataset
    duplicates = 5
    saved_dataset.add_source("file1", TextFile)
    saved_dataset.add_source("file2", TextFile)
    saved_dataset.add_sink("concatenated", TextFile)
    saved_dataset.apply(
        name="a_pipeline",
        workflow=concatenate_task(name="workflow", duplicates=duplicates),
        inputs=[("file1", "in_file1"), ("file2", "in_file2")],
        outputs=[("concatenated", "out_file")],
    )
    # Add source column to saved dataset
    result = cli_runner(
        apply,
        [
            saved_dataset.locator,
            "a_pipeline",
            "frametree.testing.tasks:" + concatenate_task.__name__,
            "--source",
            "file1",
            "in_file1",
            "text/text-file",
            "--source",
            "file2",
            "in_file2",
            "text/text-file",
            "--sink",
            "concatenated",
            "out_file",
            "text/text-file",
            "--parameter",
            "duplicates",
            str(duplicates),
        ],
    )
    assert result.exit_code == 0, show_cli_trace(result)
    loaded_dataset = FrameSet.load(saved_dataset.locator)
    assert saved_dataset.pipelines == loaded_dataset.pipelines


def test_derive_cli(saved_dataset, concatenate_task, cli_runner):
    # Get CLI name for dataset (i.e. file system path prepended by 'file//')
    bp = saved_dataset.__annotations__["blueprint"]
    duplicates = 3
    # Start generating the arguments for the CLI
    # Add source to loaded dataset
    result = cli_runner(
        apply,
        [
            saved_dataset.locator,
            "a_pipeline",
            "frametree.testing.tasks:" + concatenate_task.__name__,
            "--source",
            "file1",
            "in_file1",
            "text/text-file",
            "--source",
            "file2",
            "in_file2",
            "text/text-file",
            "--sink",
            "concatenated",
            "out_file",
            "text/text-file",
            "--parameter",
            "duplicates",
            str(duplicates),
        ],
    )
    assert result.exit_code == 0, show_cli_trace(result)
    # Add source column to saved dataset
    result = cli_runner(
        derive, [saved_dataset.locator, "concatenated", "--plugin", "serial"]
    )
    assert result.exit_code == 0, show_cli_trace(result)
    sink = saved_dataset.add_sink("concatenated", TextFile)
    assert len(sink) == reduce(mul, bp.dim_lengths)
    fnames = ["file1.txt", "file2.txt"]
    if concatenate_task.__name__.endswith("reverse"):
        fnames = [f[::-1] for f in fnames]
    expected_contents = "\n".join(fnames * duplicates)
    for item in sink:
        with open(item) as f:
            contents = f.read()
        assert contents == expected_contents


LICENSE_CONTENTS = "test license"


@pytest.fixture(scope="module")
def test_license():
    tmp_dir = Path(tempfile.mkdtemp())
    test_license = tmp_dir / "license.txt"
    test_license.write_text(LICENSE_CONTENTS)
    return str(test_license)


def test_cli_install_dataset_license(
    saved_dataset: FrameSet, test_license, frametree_home, cli_runner, tmp_path
):
    store_nickname = saved_dataset.id + "_store"
    license_name = "test-license"
    saved_dataset.store.save(store_nickname)
    address = store_nickname + "//" + saved_dataset.id + "@" + saved_dataset.name

    result = cli_runner(
        install_license,
        [
            license_name,
            test_license,
            address,
        ],
    )
    assert result.exit_code == 0, show_cli_trace(result)
    assert saved_dataset.get_license_file(license_name).contents == LICENSE_CONTENTS

    # Test overwriting
    new_contents = "new_contents"
    new_license = tmp_path / "new-license.txt"
    new_license.write_text(new_contents)

    result = cli_runner(
        install_license,
        [
            license_name,
            str(new_license),
            address,
        ],
    )
    assert result.exit_code == 0, show_cli_trace(result)
    assert saved_dataset.get_license_file(license_name).contents == new_contents


def test_cli_install_site_license(
    data_store,
    test_license: str,
    frametree_home,
    cli_runner,
):
    store_nickname = "site_license_store"
    license_name = "test-license"
    data_store.save(store_nickname)

    if isinstance(data_store, MockRemote):
        env = {
            data_store.SITE_LICENSES_USER_ENV: "arbitrary_user",
            data_store.SITE_LICENSES_PASS_ENV: "arbitrary_password",
        }
    else:
        env = {}

    result = cli_runner(
        install_license,
        [
            license_name,
            test_license,
            store_nickname,
        ],
        env=env,
    )

    assert result.exit_code == 0, show_cli_trace(result)

    assert (
        data_store.get_site_license_file(
            license_name, user="arbitrary_user", password="arbitrary_password"
        ).contents
        == LICENSE_CONTENTS
    )



from frametree.common import Clinical


def test_is_parent():
    assert not Clinical.session.is_parent(Clinical.session)
    assert not Clinical.session.is_parent(Clinical.subject)
    assert not Clinical.session.is_parent(Clinical.member)
    assert not Clinical.session.is_parent(Clinical.group)
    assert not Clinical.session.is_parent(Clinical.visit)
    assert not Clinical.session.is_parent(Clinical.groupedvisit)
    assert not Clinical.session.is_parent(Clinical.matchedvisit)
    assert not Clinical.session.is_parent(Clinical.constant)

    assert Clinical.member.is_parent(Clinical.session)
    assert Clinical.member.is_parent(Clinical.subject)
    assert not Clinical.member.is_parent(Clinical.member)
    assert not Clinical.member.is_parent(Clinical.group)
    assert not Clinical.member.is_parent(Clinical.visit)
    assert not Clinical.member.is_parent(Clinical.groupedvisit)
    assert Clinical.member.is_parent(Clinical.matchedvisit)
    assert not Clinical.member.is_parent(Clinical.constant)

    assert Clinical.group.is_parent(Clinical.session)
    assert Clinical.group.is_parent(Clinical.subject)
    assert not Clinical.group.is_parent(Clinical.member)
    assert not Clinical.group.is_parent(Clinical.group)
    assert not Clinical.group.is_parent(Clinical.visit)
    assert Clinical.group.is_parent(Clinical.groupedvisit)
    assert not Clinical.group.is_parent(Clinical.matchedvisit)
    assert not Clinical.group.is_parent(Clinical.constant)

    assert Clinical.visit.is_parent(Clinical.session)
    assert not Clinical.visit.is_parent(Clinical.subject)
    assert not Clinical.visit.is_parent(Clinical.member)
    assert not Clinical.visit.is_parent(Clinical.group)
    assert not Clinical.visit.is_parent(Clinical.visit)
    assert Clinical.visit.is_parent(Clinical.groupedvisit)
    assert Clinical.visit.is_parent(Clinical.matchedvisit)
    assert not Clinical.visit.is_parent(Clinical.constant)

    assert Clinical.subject.is_parent(Clinical.session)
    assert not Clinical.subject.is_parent(Clinical.subject)
    assert not Clinical.subject.is_parent(Clinical.member)
    assert not Clinical.subject.is_parent(Clinical.group)
    assert not Clinical.subject.is_parent(Clinical.visit)
    assert not Clinical.subject.is_parent(Clinical.groupedvisit)
    assert not Clinical.subject.is_parent(Clinical.matchedvisit)
    assert not Clinical.subject.is_parent(Clinical.constant)

    assert Clinical.groupedvisit.is_parent(Clinical.session)
    assert not Clinical.groupedvisit.is_parent(Clinical.subject)
    assert not Clinical.groupedvisit.is_parent(Clinical.member)
    assert not Clinical.groupedvisit.is_parent(Clinical.group)
    assert not Clinical.groupedvisit.is_parent(Clinical.visit)
    assert not Clinical.groupedvisit.is_parent(Clinical.groupedvisit)
    assert not Clinical.groupedvisit.is_parent(Clinical.matchedvisit)
    assert not Clinical.groupedvisit.is_parent(Clinical.constant)

    assert Clinical.matchedvisit.is_parent(Clinical.session)
    assert not Clinical.matchedvisit.is_parent(Clinical.subject)
    assert not Clinical.matchedvisit.is_parent(Clinical.member)
    assert not Clinical.matchedvisit.is_parent(Clinical.group)
    assert not Clinical.matchedvisit.is_parent(Clinical.visit)
    assert not Clinical.matchedvisit.is_parent(Clinical.groupedvisit)
    assert not Clinical.matchedvisit.is_parent(Clinical.matchedvisit)
    assert not Clinical.matchedvisit.is_parent(Clinical.constant)

    assert Clinical.constant.is_parent(Clinical.session)
    assert Clinical.constant.is_parent(Clinical.subject)
    assert Clinical.constant.is_parent(Clinical.member)
    assert Clinical.constant.is_parent(Clinical.group)
    assert Clinical.constant.is_parent(Clinical.visit)
    assert Clinical.constant.is_parent(Clinical.groupedvisit)
    assert Clinical.constant.is_parent(Clinical.matchedvisit)
    assert not Clinical.constant.is_parent(Clinical.constant)



from __future__ import annotations
import typing as ty
import attrs
from fileformats.core import DataType
from frametree.core.exceptions import FrameTreeError

if ty.TYPE_CHECKING:  # pragma: no cover
    from .row import DataRow
    from .column import DataColumn
    from .entry import DataEntry


@attrs.define
class DataCell:
    """
    An abstract container representing the intersection between a row and a column,
    which points to an item within a dataset.

    Parameters
    ----------
    row : DataRow
        the row the cell belongs to
    column : DataColumn
        the column the cell belongs to
    entry : DataEntry or None
        the entry in the row that matches the column
    provenance : Provenance | None
        The recorded provenance for the item stored within the data cell,
        if applicable
    """

    row: DataRow
    column: DataColumn
    entry: DataEntry
    provenance: ty.Dict[str, ty.Any] = None

    @property
    def datatype(self):
        return self.column.datatype

    @property
    def is_empty(self):
        return self.entry is None

    @property
    def item(self) -> DataType:
        if self.is_empty:
            raise FrameTreeError(f"Cannot access item of empty cell, {self}")
        return self.entry.get_item(self.column.datatype)

    @item.setter
    def item(self, item: DataType):
        if not self.column.is_sink:
            raise FrameTreeError(
                f"Cannot set data items ({item}) into source column cell {self}"
            )
        item = self.datatype(item)
        if self.is_empty:
            entry = self.row.frameset.store.post(
                item=item, path=self.column.path, datatype=self.datatype, row=self.row
            )
            self.entry = entry
        else:
            self.entry.item = item

    @classmethod
    def intersection(
        cls, column: DataColumn, row: DataRow, allow_empty: ty.Optional[bool] = None
    ) -> DataCell:
        if allow_empty is None:
            allow_empty = column.is_sink
        return cls(
            row=row,
            column=column,
            entry=column.match_entry(row, allow_none=allow_empty),
        )



from frametree.core.packaging import package_from_module
from frametree.core.utils import path2varname, varname2path


def test_package_from_module():
    assert package_from_module("frametree.core").key == "frametree"
    assert package_from_module("pydra.engine").key == "pydra"


def test_path2varname():
    escape_pairs = [
        ("dwi/dir-LR_dwi", "dwi__l__dir__H__LR_u_dwi"),
        ("func/task-rest_bold", "func__l__task__H__rest_u_bold"),
        (
            "with spaces and__ underscores",
            "with__s__spaces__s__and_u__u___s__underscores",
        ),
        ("__a.very$illy*ath~", "XXX_u__u_a__o__very__dollar__illy__star__ath__tilde__"),
        ("anat/T1w", "anat__l__T1w"),
        ("anat__l__T1w", "anat_u__u_l_u__u_T1w"),
        ("_u__u_", "XXX_u_u_u__u_u_u_"),
    ]
    for path, varname in escape_pairs:
        assert path2varname(path) == varname
        assert varname2path(varname) == path
        assert varname2path(varname2path(path2varname(path2varname(path)))) == path



from __future__ import annotations
import typing as ty
import attrs
from fileformats.core import DataType
from frametree.core.exceptions import FrameTreeDataMatchError, FrameTreeUsageError
from .quality import DataQuality

if ty.TYPE_CHECKING:  # pragma: no cover
    from .row import DataRow


@attrs.define
class ItemMetadata:
    """Metadata that is either manually set at initialisation of the DataEntry (if
    easily extracted from the data store), or lazily loaded from the entry's item if the
    entry datatype"""

    loaded: dict = attrs.field(
        default=None, converter=lambda x: dict(x) if x is not None else {}
    )
    _entry: DataEntry = attrs.field(default=None, init=False, repr=False)
    _has_been_loaded: bool = attrs.field(default=False, init=False, repr=False)

    def __iter__(self):
        raise NotImplementedError

    def __getitem__(self, key):
        try:
            return self.loaded[key]
        except KeyError:
            if not self._has_been_loaded:
                self.load()
        return self.loaded[key]

    def load(self, overwrite=False):
        assert self._entry is not None
        item_metadata = self._entry.item.metadata
        if not overwrite:
            mismatching = [
                k
                for k in set(self.loaded) & set(item_metadata)
                if self.loaded[k] != item_metadata[k]
            ]
            if mismatching:
                raise RuntimeError(
                    "Mismatch in values between loaded and loaded metadata values, "
                    "use 'load(overwrite=True)' to overwrite:\n"
                    + "\n".join(
                        f"{k}: loaded={self.loaded[k]}, loaded={item_metadata[k]}"
                        for k in mismatching
                    )
                )
        self.loaded.update(item_metadata)
        self._has_been_loaded = True


@attrs.define()
class DataEntry:
    """An entry in a node of the dataset tree, such as a scan in an imaging
    session in a "session node" or group-level derivative in a "group node"

    Parameters
    ----------
    id : str
        the ID of the entry within the node
    datatype : type (subclass of DataType)
        the type of the data entry
    uri : str, optional
        a URI uniquely identifying the data entry
    item_metadata : dict[str, Any]
        metadata associated with the data item itself (e.g. pulled from a file header).
        Can be supplied either when the entry is initialised (i.e. from previously extracted
        fields stored within the data store), or read from the item itself.
    order : int, optional
        the order in which the entry appears in the node (where applicable)
    provenance : dict, optional
        the provenance associated with the derivation of the entry by FrameTree
        (only applicable to derivatives not source data)
    checksums : dict[str, str], optional
        checksums for all of the files in the data entry
    """

    path: str = attrs.field()
    datatype: type
    row: DataRow
    uri: str
    item_metadata: ItemMetadata = attrs.field(
        default=None, converter=ItemMetadata, repr=False, kw_only=True
    )
    order: ty.Optional[int] = None
    quality: DataQuality = DataQuality.usable
    checksums: ty.Dict[str, ty.Union[str, dict]] = attrs.field(
        default=None, repr=False, eq=False
    )

    @path.validator
    def path_validator(self, _, path: str):
        path, dataset_name = self.split_dataset_name_from_path(path)
        if dataset_name and not dataset_name.isidentifier():
            raise FrameTreeUsageError(
                f"Path '{path}' has an invalid dataset_name '{dataset_name}')"
            )

    def __attrs_post_init__(self):
        self.item_metadata._entry = self

    @property
    def item(self) -> DataType:
        return self.get_item()

    @item.setter
    def item(self, item):
        if isinstance(item, DataType):
            if not isinstance(item, self.datatype):
                raise FrameTreeDataMatchError(
                    f"Cannot put {item} into {self.datatype} entry of {self.row}"
                )
        else:
            item = self.datatype(item)
        self.row.frameset.store.put(self, item)

    def get_item(self, datatype=None):
        if datatype is None:
            datatype = self.datatype
        return self.row.frameset.store.get(self, datatype)

    @property
    def recorded_checksums(self):
        if self.provenance is None:
            return None
        else:
            return self.provenance.outputs[self.path]

    @property
    def is_derivative(self):
        return self.path_is_derivative(self.path)

    @property
    def base_path(self):
        return self.split_dataset_name_from_path(self.path)[0]

    @property
    def dataset_name(self):
        return self.split_dataset_name_from_path(self.path)[1]

    @classmethod
    def split_dataset_name_from_path(cls, path):
        parts = path.split("@")
        if len(parts) == 1:
            dataset_name = None
        else:
            path, dataset_name = parts
        if len(parts) > 2:
            raise FrameTreeUsageError(
                f"Entry paths can't have more than one '@' symbol, given {path})"
            )
        return path, dataset_name

    @classmethod
    def path_is_derivative(cls, path):
        return cls.split_dataset_name_from_path(path)[1] is not None



from .store import store  # noqa: F401
from .frameset import (  # noqa: F401
    define,
    add_source,
    add_sink,
    missing_items,
    export,
    copy,
)
from .processing import derive, apply, install_license  # noqa: F401
